/* FILE: twosurfaceexchange.cc            -*-Mode: c++-*-
 *
 * Long range (RKKY-style) exchange energy on links across spacer
 * layer, with surface (interfacial) exchange energy coefficents
 * sigma (bilinear) and sigma2 (biquadratic).
 *
 */

#include "nb.h"
#include "director.h"
#include "mesh.h"
#include "meshvalue.h"
#include "oxswarn.h"
#include "simstate.h"
#include "threevector.h"
#include "rectangularmesh.h"
#include "twosurfaceexchange.h"
#include "energy.h"             // Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_TwoSurfaceExchange);

/* End includes */

// Revision information, set via CVS keyword substitution
static const Oxs_WarningMessageRevisionInfo revision_info
  (__FILE__,
   "$Revision: 1.14 $",
   "$Date: 2009-10-12 06:34:49 $",
   "$Author: donahue $",
   "Michael J. Donahue (michael.donahue@nist.gov)");


// Constructor
Oxs_TwoSurfaceExchange::Oxs_TwoSurfaceExchange(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,argstr),
    mesh_id(0),report_max_spin_angle(0)
{
  // Process arguments
  init_sigma = GetRealInitValue("sigma",0.);
  init_sigma2 = GetRealInitValue("sigma2",0.);

  // First boundary.  Import parameter string should be a 10 element
  // list of name-value pairs:
  //         NAME         VALUE
  //         atlas        atlas reference
  //         region       region name
  //         scalarfield  scalar field reference
  //         scalarvalue  scalar field level surface reference value
  //         scalarside   either "+" or "-", denoting inner surface side
  CheckInitValueParamCount("surface1",10);
  vector<String> surface1;
  FindRequiredInitValue("surface1",surface1);
  BOOL badinput1=0;
  String error1;
  if(surface1[0].compare("atlas")!=0) {
    badinput1 = 1;
    error1 = String("Bad name \"") + surface1[0]
      + String("\"; should be \"atlas\"");
  } else if(surface1[2].compare("region")      != 0) {
    badinput1 = 1;
    error1 = String("Bad name \"") + surface1[2]
      + String("\"; should be \"region\"");
  } else if(surface1[4].compare("scalarfield") != 0) {
    badinput1 = 1;
    error1 = String("Bad name \"") + surface1[4]
      + String("\"; should be \"scalarfield\"");
  } else if(surface1[6].compare("scalarvalue") != 0) {
    badinput1 = 1;
    error1 = String("Bad name \"") + surface1[6]
      + String("\"; should be \"scalarvalue\"");
  } else if(surface1[8].compare("scalarside")  != 0) {
    badinput1 = 1;
    error1 = String("Bad name \"") + surface1[8]
      + String("\"; should be \"scalarside\"");
  } else {
    Oxs_SplitList param_strings;
    vector<String> params;
    param_strings.Split(surface1[1].c_str());
    param_strings.FillParams(params);
    OXS_GET_EXT_OBJECT(params,Oxs_Atlas,atlas1);
    region1 = surface1[3];
    param_strings.Split(surface1[5].c_str());
    param_strings.FillParams(params);
    OXS_GET_EXT_OBJECT(params,Oxs_ScalarField,bdry1);
    bdry1_value = Nb_Atof(surface1[7].c_str());
    bdry1_side = surface1[9];
    if(atlas1->GetRegionId(region1) == -1) {
      badinput1 = 1;
      error1 = String("Unrecognized region \"")
        + region1 + String("\"");
    } else if(bdry1_side.compare("+")!=0 && bdry1_side.compare("-")!=0) {
      badinput1 = 1;
      error1 = String("Bad scalarside string \"")
        + bdry1_side + String("\"; should be \"+\" or \"-\"");
    }
  }
  DeleteInitValue("surface1");

  // Second boundary.  Import parameter string same format
  // as for first boundary.
  CheckInitValueParamCount("surface2",10);
  vector<String> surface2;
  FindRequiredInitValue("surface2",surface2);
  BOOL badinput2=0;
  String error2;
  if(surface2[0].compare("atlas")!=0) {
    badinput2 = 1;
    error2 = String("Bad name \"") + surface2[0]
      + String("\"; should be \"atlas\"");
  } else if(surface2[2].compare("region")      != 0) {
    badinput2 = 1;
    error2 = String("Bad name \"") + surface2[2]
      + String("\"; should be \"region\"");
  } else if(surface2[4].compare("scalarfield") != 0) {
    badinput2 = 1;
    error2 = String("Bad name \"") + surface2[4]
      + String("\"; should be \"scalarfield\"");
  } else if(surface2[6].compare("scalarvalue") != 0) {
    badinput2 = 1;
    error2 = String("Bad name \"") + surface2[6]
      + String("\"; should be \"scalarvalue\"");
  } else if(surface2[8].compare("scalarside")  != 0) {
    badinput2 = 1;
    error2 = String("Bad name \"") + surface2[8]
      + String("\"; should be \"scalarside\"");
  } else {
    Oxs_SplitList param_strings;
    vector<String> params;
    param_strings.Split(surface2[1].c_str());
    param_strings.FillParams(params);
    OXS_GET_EXT_OBJECT(params,Oxs_Atlas,atlas2);
    region2 = surface2[3];
    param_strings.Split(surface2[5].c_str());
    param_strings.FillParams(params);
    OXS_GET_EXT_OBJECT(params,Oxs_ScalarField,bdry2);
    bdry2_value = Nb_Atof(surface2[7].c_str());
    bdry2_side = surface2[9];
    if(atlas2->GetRegionId(region2) == -1) {
      badinput2 = 1;
      error2 = String("Unrecognized region \"")
        + region2 + String("\"");
    } else if(bdry2_side.compare("+")!=0 && bdry2_side.compare("-")!=0) {
      badinput2 = 1;
      error2 = String("Bad scalarside string \"")
        + bdry2_side + String("\"; should be \"+\" or \"-\"");
    }
  }
  DeleteInitValue("surface2");

  if(badinput1 || badinput2) {
    String msg=
      String("Bad surface parameter block, inside "
             "Oxs_TwoSurfaceExchange mif Specify block for object ")
      +  String(InstanceName()) + String(":");
    if(badinput1) msg += String(" surface1 block is bad: ") + error1;
    if(badinput1 && badinput2) msg += String(", ");
    if(badinput2) msg += String(" surface2 block is bad: ") + error2;
    msg += String(".");
    throw Oxs_ExtError(msg.c_str());
  }

  // Report max spin angle?  Generally this value is used to detect
  // grids that are too coarse, which is not applicable in this instance
  // because the connected spins are assumed to be spatially separated.
  // This output is provided in case it proves useful, but is disabled
  // by default.
  if(GetIntInitValue("report_max_spin_angle",0)) {
    report_max_spin_angle = 1;
  } else {
    report_max_spin_angle = 0;
  }

  VerifyAllInitArgsUsed();

  // Setup outputs
  if(report_max_spin_angle) {
    maxspinangle_output.Setup(this,InstanceName(),"Max Spin Ang","deg",1,
                              &Oxs_TwoSurfaceExchange::UpdateDerivedOutputs);
    maxspinangle_output.Register(director,0);
    stage_maxspinangle_output.Setup(this,InstanceName(),"Stage Max Spin Ang","deg",1,
                                    &Oxs_TwoSurfaceExchange::UpdateDerivedOutputs);
    stage_maxspinangle_output.Register(director,0);
    run_maxspinangle_output.Setup(this,InstanceName(),"Run Max Spin Ang","deg",1,
                                  &Oxs_TwoSurfaceExchange::UpdateDerivedOutputs);
    run_maxspinangle_output.Register(director,0);
  }
}

Oxs_TwoSurfaceExchange::~Oxs_TwoSurfaceExchange()
{}

BOOL Oxs_TwoSurfaceExchange::Init()
{
  mesh_id = 0;
  links.clear();
  return Oxs_Energy::Init();
}

void Oxs_TwoSurfaceExchange::FillLinkList
(const Oxs_RectangularMesh* mesh) const
{ // Queries the mesh with each surface spec,
  // and then pairs up cells between surfaces.
  // Each cell on the first surface is paired up with
  // the closest cell from the second surface.  Note
  // the asymmetry in this pairing procedure---generally,
  // one will probably want to make the smaller surface
  // the "first" surface.  Consider these pictures:
  //
  //          22222222222222 <-- Surface 2
  //              ||||||           
  //              ||||||  <-- Links
  //              ||||||         
  //              111111  <-- Surface 1
  //
  //  as compared to
  //
  //          11111111111111 <-- Surface 1
  //           \\\||||||///        
  //            \\||||||// <-- Links
  //             \||||||/        
  //              222222  <-- Surface 2
  //
  // In the second picture, all the outer cells in the top
  // surface get paired up with the outer elements in the
  // lower surface.

  // Get cell lists for each surface
  vector<UINT4m> bdry1_cells,bdry2_cells;
  mesh->BoundaryList(*atlas1,region1,*bdry1,bdry1_value,bdry1_side,
                     bdry1_cells);
  mesh->BoundaryList(*atlas2,region2,*bdry2,bdry2_value,bdry2_side,
                     bdry2_cells);
  if(bdry1_cells.empty() || bdry2_cells.empty()) {
    String msg =
      String("Empty bdry/surface list with current mesh"
             " detected in Oxs_TwoSurfaceExchange::FillLinkList()"
             " routine of object ") + String(InstanceName());
    throw Oxs_ExtError(msg.c_str());
  }

  // For each cell in bdry1_cells, find closest cell in bdry2_cells,
  // and store this pair in "links".
  links.clear();  // Throw away previous links, if any.
  vector<UINT4m>::iterator it1 = bdry1_cells.begin();
  while(it1!=bdry1_cells.end()) {
    ThreeVector cell1;
    mesh->Center(*it1,cell1);
    vector<UINT4m>::iterator it2 = bdry2_cells.begin();
    ThreeVector vtemp;
    mesh->Center(*it2,vtemp);
    vtemp -= cell1;
    REAL8m min_distsq = vtemp.MagSq();
    UINT4m min_index = *it2;
    while( (++it2) != bdry2_cells.end()) {
      mesh->Center(*it2,vtemp);
      vtemp -= cell1;
      REAL8m distsq = vtemp.MagSq();
      if(distsq<min_distsq) {
        min_distsq = distsq;
        min_index = *it2;
      }
    }
    if(min_distsq>0.) {
      // The sigma coef's need to be weighted by the inverse
      // of the thickness of the compuational cell in the
      // direction of the link.  This is necessary so that
      // the total surface (interfacial) exchange energy is
      // independent of the discretization size, as the
      // discretization size is reduces to 0.  (As the
      // computational cell size is reduced, the surface
      // energy is squeezed into an ever thinner layer of
      // cells along the surface.)
      ThreeVector wtemp;
      mesh->Center(min_index,wtemp);
      wtemp -= cell1;
      REAL8m cellwgt = wtemp.x*mesh->EdgeLengthX()
                     + wtemp.y*mesh->EdgeLengthY()
                     + wtemp.z*mesh->EdgeLengthZ();
      if(cellwgt>0.) { // Safety
	cellwgt = sqrt(min_distsq)/cellwgt;
	// min_distsq _should_ be wtemp.MagSq(), so
	// cellwgt is |wtemp|/<wtemp,celldims>, i.e.,
	// < wtemp/|wtemp| , celldims >^{-1}
      }

      Oxs_TwoSurfaceExchangeLinkParams ltemp;
      ltemp.index1 = *it1;    ltemp.index2 = min_index;
      ltemp.exch_coef1 = (init_sigma + init_sigma2) * cellwgt;
      ltemp.exch_coef2 = -init_sigma2 * cellwgt;
      // NB: The init_sigma and init_sigma2 references above
      //  will likely be replaced in the future with a call
      //  to a user-defined function taking as imports the
      //  center locations of the two cells in the link pair.

      links.push_back(ltemp);
    }
    ++it1;
  }
}

void Oxs_TwoSurfaceExchange::GetEnergy
(const Oxs_SimState& state,
 Oxs_EnergyData& oed
 ) const
{
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);

  // Use supplied buffer space, and reflect that use in oed.
  oed.energy = oed.energy_buffer;
  oed.field = oed.field_buffer;
  Oxs_MeshValue<REAL8m>& energy = *oed.energy_buffer;
  Oxs_MeshValue<ThreeVector>& field = *oed.field_buffer;

  const Oxs_RectangularMesh* mesh
    = dynamic_cast<const Oxs_RectangularMesh*>(state.mesh);
  if(mesh==NULL) {
    String msg =
      String("Import mesh to Oxs_TwoSurfaceExchange::GetEnergy()"
             " routine of object ") + String(InstanceName())
      + String(" is not an Oxs_RectangularMesh object.");
    throw Oxs_ExtError(msg.c_str());
  }
  // If mesh has changed, re-pick link selections
  if(mesh_id !=  mesh->Id()) {
    FillLinkList(mesh);
    mesh_id=mesh->Id();
  }

  // Zero entire energy and field meshes
  UINT4m size=mesh->Size();
  UINT4m index;
  for(index=0;index<size;index++) energy[index]=0.;
  for(index=0;index<size;index++) field[index].Set(0.,0.,0.);

  // Iterate through link list and accumulate energies and fields
  REAL8m mindot = 1;
  REAL8m hcoef = 2.0/MU0;
  vector<Oxs_TwoSurfaceExchangeLinkParams>::const_iterator it
    = links.begin();
  for(it=links.begin();it!=links.end();++it) {
    UINT4m i = it->index1;
    UINT4m j = it->index2;
    if(Ms_inverse[i]==0. || Ms_inverse[j]==0.) continue; // Ms=0; skip
    REAL8m ecoef1 = it->exch_coef1;
    REAL8m ecoef2 = it->exch_coef2;
    ThreeVector mdiff = spin[i]-spin[j];
    REAL8m dot = spin[i] * spin[j];
    if(dot<mindot) mindot = dot;
    REAL8m temp = 0.5*mdiff.MagSq();
    REAL8m elink = (ecoef1 + ecoef2*temp)*temp; // Energy density
    energy[i] += elink;
    energy[j] += elink;
    mdiff *= hcoef*(ecoef1+2*ecoef2*temp);
    field[i] += -1*Ms_inverse[i]*mdiff;
    field[j] +=    Ms_inverse[j]*mdiff;
  }

  // Set maxang data
  REAL8m maxang;
  // Handle extremal cases separately in case spins drift off S2
  if(mindot >  1 - 8*REAL8_EPSILON)      maxang = 0.0;
  else if(mindot < -1 + 8*REAL8_EPSILON) maxang = 180.0;
  else                                   maxang = acos(mindot)*(180.0/PI);

  REAL8m dummy_value;
  String msa_name = MaxSpinAngleStateName();
  if(state.GetDerivedData(msa_name,dummy_value)) {
    // Ideally, energy values would never be computed more than once
    // for any one state, but in practice it seems inevitable that
    // such will occur on occasion.  For example, suppose a user
    // requests output on a state obtained by a stage crossing (as
    // opposed to a state obtained through a normal intrastage step);
    // a subsequent ::Step operation will re-compute the energies
    // because not all the information needed by the step transition
    // machinery is cached from an energy computation.  Even user
    // output requests on post-::Step states is problematic if some of
    // the requested output is not cached as part of the step
    // proceedings.  A warning is put into place below for debugging
    // purposes, but in general an error is raised only if results
    // from the recomputation are different than originally.
#ifndef NDEBUG
    static Oxs_WarningMessage maxangleset(3);
    maxangleset.Send(revision_info,OC_STRINGIFY(__LINE__),
                     "Programming error?"
                     " Oxs_TwoSurfaceExchange max spin angle set twice.");
#endif
    if(fabs(dummy_value-maxang)>8*180.*REAL8_EPSILON) {
      char errbuf[1024];
      Oc_Snprintf(errbuf,sizeof(errbuf),
                  "Programming error:"
                  " Oxs_TwoSurfaceExchange max spin angle set to"
                  " two different values;"
                  " orig val=%#.17g, new val=%#.17g",
                  dummy_value,maxang);
      throw Oxs_ExtError(this,errbuf);
    }
  } else {
    state.AddDerivedData(msa_name,maxang);
  }

  // Run and stage angle data depend on data from the previous state.
  // In the case that the energy (and hence max stage and run angle)
  // for the current state was computed previously, then the previous
  // state may have been dropped.  So, compute and save run and stage
  // angle data iff they are not already computed.

  // Check stage and run max angle data from previous state
  const Oxs_SimState* oldstate = NULL;
  REAL8m stage_maxang = -1;
  REAL8m run_maxang = -1;
  String smsa_name = StageMaxSpinAngleStateName();
  String rmsa_name = RunMaxSpinAngleStateName();
  if(state.previous_state_id &&
     (oldstate
      = director->FindExistingSimulationState(state.previous_state_id)) ) {
    if(oldstate->stage_number != state.stage_number) {
      stage_maxang = 0.0;
    } else {
      if(oldstate->GetDerivedData(smsa_name,dummy_value)) {
        stage_maxang = dummy_value;
      }
    }
    if(oldstate->GetDerivedData(rmsa_name,dummy_value)) {
      run_maxang = dummy_value;
    }
  }
  if(stage_maxang<maxang) stage_maxang = maxang;
  if(run_maxang<maxang)   run_maxang = maxang;

  // Stage max angle data
  if(!state.GetDerivedData(smsa_name,dummy_value)) {
    state.AddDerivedData(smsa_name,stage_maxang);
  }

  // Run max angle data
  if(!state.GetDerivedData(rmsa_name,dummy_value)) {
    state.AddDerivedData(rmsa_name,run_maxang);
  }
}

void Oxs_TwoSurfaceExchange::UpdateDerivedOutputs(const Oxs_SimState& state)
{
  maxspinangle_output.cache.state_id
    = stage_maxspinangle_output.cache.state_id
    = run_maxspinangle_output.cache.state_id
    = 0;  // Mark change in progress

  String dummy_name = MaxSpinAngleStateName();
  if(!state.GetDerivedData(dummy_name.c_str(),
                           maxspinangle_output.cache.value)) {
    // Error; This should always be set.  For now, just set the value to
    // -1, but in the future should consider throwing an exception.
    maxspinangle_output.cache.value = -1.0;
  }

  String stage_dummy_name = StageMaxSpinAngleStateName();
  if(!state.GetDerivedData(stage_dummy_name.c_str(),
                           stage_maxspinangle_output.cache.value)) {
    // Error; This should always be set.  For now, just set the value to
    // -1, but in the future should consider throwing an exception.
    stage_maxspinangle_output.cache.value = -1.0;
  }

  String run_dummy_name = RunMaxSpinAngleStateName();
  if(!state.GetDerivedData(run_dummy_name.c_str(),
                           run_maxspinangle_output.cache.value)) {
    // Error; This should always be set.  For now, just set the value to
    // -1, but in the future should consider throwing an exception.
    run_maxspinangle_output.cache.value = -1.0;
  }

  maxspinangle_output.cache.state_id
    = stage_maxspinangle_output.cache.state_id
    = run_maxspinangle_output.cache.state_id
    = state.Id();
}
