/* FILE: hexanisotropy.cc            -*-Mode: c++-*-
 *
 * Hex Anisotropy, derived from Oxs_Energy class.
 *
 */

#include <string>

#include "oc.h"
#include "nb.h"
#include "threevector.h"
#include "director.h"
#include "simstate.h"
#include "ext.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "hexanisotropy.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

OC_USE_STRING;

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_HexAnisotropy);

/* End includes */


// Constructor
Oxs_HexAnisotropy::Oxs_HexAnisotropy(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,argstr), mesh_id(0)
{
  // Process arguments
  OXS_GET_INIT_EXT_OBJECT("K1",Oxs_ScalarField,K1_init);
  OXS_GET_INIT_EXT_OBJECT("K2",Oxs_ScalarField,K2_init);
  OXS_GET_INIT_EXT_OBJECT("K3",Oxs_ScalarField,K3_init);
  OXS_GET_INIT_EXT_OBJECT("K4",Oxs_ScalarField,K4_init);
  OXS_GET_INIT_EXT_OBJECT("axis1",Oxs_VectorField,axis1_init);
  OXS_GET_INIT_EXT_OBJECT("axis2",Oxs_VectorField,axis2_init);
  VerifyAllInitArgsUsed();
}

BOOL Oxs_HexAnisotropy::Init()
{
  mesh_id = 0;
  K1.Release();
  K2.Release();
  K3.Release();
  K4.Release();
  axis1.Release();
  axis2.Release();
  return Oxs_Energy::Init();
}

void Oxs_HexAnisotropy::GetEnergy
(const Oxs_SimState& state,
 Oxs_EnergyData& oed
 ) const
{
  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;

  // Use supplied buffer space, and reflect that use in oed.
  oed.energy = oed.energy_buffer;
  oed.field = oed.field_buffer;
  Oxs_MeshValue<REAL8m>& energy = *oed.energy_buffer;
  Oxs_MeshValue<ThreeVector>& field = *oed.field_buffer;

  UINT4m size = state.mesh->Size();
  if(mesh_id != state.mesh->Id()) {
    // This is either the first pass through, or else mesh
    // has changed.
    mesh_id=0;
    K1_init->FillMeshValue(state.mesh,K1);
    K2_init->FillMeshValue(state.mesh,K2);
    K3_init->FillMeshValue(state.mesh,K3);
    K4_init->FillMeshValue(state.mesh,K4);
    axis1_init->FillMeshValue(state.mesh,axis1);
    axis2_init->FillMeshValueOrthogonal(state.mesh,axis1,axis2);
    for(UINT4m i=0;i<size;i++) {
      // Much of the code below requires axis1 and axis2 to be
      // orthogonal unit vectors.  Guarantee this is the case:
      const REAL8m eps = 1e-14;
      if(fabs(axis1[i].MagSq()-1)>eps) {
	String msg = String("Invalid initialization detected for object ")
	  + String(InstanceName())
	  + String(": Anisotropy axis 1 isn't norm 1");
	throw Oxs_ExtError(msg.c_str());
      }
      if(fabs(axis2[i].MagSq()-1)>eps) {
	String msg = String("Invalid initialization detected for object ")
	  + String(InstanceName())
	  + String(": Anisotropy axis 2 isn't norm 1");
	throw Oxs_ExtError(msg.c_str());
      }
      if(fabs(axis1[i]*axis2[i])>eps) {
	String msg = String("Invalid initialization detected for object ")
	  + String(InstanceName())
	  + String(": Specified anisotropy axes aren't perpendicular");
	throw Oxs_ExtError(msg.c_str());
      }
    }
    mesh_id=state.mesh->Id();
  }


  for(UINT4m i=0;i<size;++i) {
    REAL8m field_mult = Ms_inverse[i]/MU0;
    if(field_mult==0.0) {
      energy[i]=0.0;
      field[i].Set(0.,0.,0.);
      continue;
    }

    const ThreeVector& u1 = axis1[i];
    const ThreeVector& u2 = axis2[i];
    const ThreeVector&  m = spin[i];
    const REAL8m k1 = K1[i];
    const REAL8m k2 = K2[i];
    const REAL8m k3 = K3[i];
    const REAL8m k4 = K4[i];

    // For hexagonal symmetry, the anisotropy energy can be written as
    //
    //      K_1 \sin^2(\theta)
    //   +  K_2 \sin^4(\theta)
    //   +  K_3 \sin^6(\theta)
    //   +  K_4 \sin^6(\theta) \cos(6\phi)
    //
    // where theta is angle between m and c-axis (axis1 := u1), and
    // phi is angle between m and basal-plane axis (axis2 := u2).
    //   Note the following identities:
    //
    //                  6            4            2
    //  cos(6a) = 32.cos (a) - 48.cos (a) + 18.cos (a) - 1
    //
    //                   5            3        
    //  sin(6a) = [32.cos (a) - 32.cos (a) + 6.cos(a)] sin(a)
    //
    // The second is used in the derivation of the field, namely
    //
    // \mu0.Ms.H = 2.cos(theta).{K1 + 2.K2.sin^2(theta)
    //                         + 3.sin^4(theta).[K3 + K4.cos(6phi)]} u1
    //           - K4.sin^6(theta).{192.cos^5(phi)
    //                               - 192.cos^3(phi) + 36.cos(phi)} u2

    ThreeVector mxu1 = m;    mxu1 ^= u1;
    REAL8m sintheta2 = mxu1.MagSq();
    REAL8m sintheta6 = sintheta2*sintheta2*sintheta2;
    REAL8m costheta = m*u1;

    REAL8m cosphi = m*u2;
    REAL8m cosphi2 = cosphi*cosphi;
    REAL8m cos6phi = ((32*cosphi2-48)*cosphi2+18)*cosphi2-1;

    energy[i] = (((k4*cos6phi+k3)*sintheta2+k2)*sintheta2+k1)*sintheta2;

    REAL8m Hcoef1 = (((k4*cos6phi+k3)*3*sintheta2+2*k2)*sintheta2+k1)
      * 2 * costheta;
    REAL8m Hcoef2 = -12*k4*cosphi*sintheta6*((cosphi2-1)*16*cosphi2+3);

    field[i] = Hcoef1*u1;
    field[i] += Hcoef2*u2;
    field[i] *= field_mult;
  }
}
