/* FILE: xpfloat.h                    -*-Mode: c++-*-
 *
 * Class for extended floating point precision using
 * compensated summation.
 * 
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 2009-07-09 03:09:27 $
 * Last modified by: $Author: donahue $
 */

#include "oc.h"

#ifndef _NB_XPFLOAT_H
#define _NB_XPFLOAT_H

class Nb_Xpfloat
{
private:
  REALWIDE x,corr;
public:
  Nb_Xpfloat(REALWIDE newx=0.0) { x=newx; corr=0.; }
  Nb_Xpfloat(const Nb_Xpfloat& y) { x=y.x; corr=y.corr; }

  void Set(REALWIDE newx) { x=newx; corr=0.; }
  Nb_Xpfloat& operator=(REALWIDE y) { Set(y); return *this; }
  Nb_Xpfloat& operator=(const Nb_Xpfloat& y) {
    x=y.x; corr=y.corr; return *this;
  }

  REALWIDE GetValue() const { return x; }
  operator REALWIDE() { return GetValue(); }

  // Defining Accum(REALWIDE) here, as opposed to in the xpfloat.cc
  // file, allows significant speedups with the g++ compiler.  It makes
  // small difference with the Intel C++ compiler (icpc), presumably
  // because of icpc's interprocedural optimization ability?
  void Accum(REALWIDE y) {
    volatile REALWIDE sum1; // Mark as volatile to protect against
    volatile REALWIDE sum2; // problems arising from extra precision.
    volatile REALWIDE corrtemp;

    // Calculate sum
    sum1=y+corr;
    sum2=x+sum1;

    // Determine new correction term.  Do in two steps, as opposed to
    // "corrtemp = (x-sum2) + sum1", because a) some compilers ignore
    // parentheses (with regards to ordering) at high optimization levels,
    // and b) we want to be sure to drop any extra precision.
    corrtemp = x - sum2;
    corrtemp += sum1;
    
    // Store results
    corr = corrtemp;
    x=sum2;
  }

  Nb_Xpfloat& operator+=(REALWIDE y) { Accum(y); return *this; }
  Nb_Xpfloat& operator-=(REALWIDE y) { Accum(-y); return *this; }
  Nb_Xpfloat& operator*=(REALWIDE y) { x*=y; corr*=y; return *this; }


  void Accum(const Nb_Xpfloat& y);
  Nb_Xpfloat& operator+=(const Nb_Xpfloat& y) { Accum(y); return *this; }
  Nb_Xpfloat& operator-=(const Nb_Xpfloat& y);
  Nb_Xpfloat& operator*=(const Nb_Xpfloat& y);

  friend Nb_Xpfloat operator*(const Nb_Xpfloat& x,const Nb_Xpfloat& y);
};

Nb_Xpfloat operator+(const Nb_Xpfloat& x,const Nb_Xpfloat& y);
Nb_Xpfloat operator-(const Nb_Xpfloat& x,const Nb_Xpfloat& y);
Nb_Xpfloat operator*(const Nb_Xpfloat& x,REALWIDE y);
inline Nb_Xpfloat operator*(REALWIDE y,const Nb_Xpfloat& x) { return x*y; }
Nb_Xpfloat operator*(const Nb_Xpfloat& x,const Nb_Xpfloat& y);

#endif // _NB_XPFLOAT_H
